@@ -18,6 +18,7 @@
 #include "ltris.h"
 #include "event.h"
 #include "chart.h"
+#include "cpu.h"
 #include "bowl.h"
 #include "shrapnells.h"
 #include "tetris.h"
@@ -319,7 +320,7 @@ int  tetris_init()
     /* create bowls according to the gametype */
     switch ( config.gametype ) {
         case GAME_DEMO:
-            bowls[0] = bowl_create( 220, 0, 490, 160, -1,-1, blocks, qmark, "Demo", 0 );
+            bowls[0] = bowl_create( 220, 0, 490, config.modern?100:160, 490,290, blocks, qmark, "Demo", 0 );
             break;
         case GAME_CLASSIC:
         case GAME_TRAINING:
@@ -653,149 +654,175 @@ void tetris_run()
     SDL_ShowCursor( 1 );
 }
 
-/*
-====================================================================
-Run a number of CPU games to get an average score gained so you'll
-see if your analyze algorithm in cpu.c cpu_analyze_bowl() sucks
-or rocks!
-====================================================================
-*/
-extern int CPU_SCORE_HOLE;
-extern int CPU_SCORE_ALT;
-extern int CPU_SCORE_LINE;
-extern int CPU_SCORE_STEEP;
-extern int CPU_SCORE_ABYSS;
-extern int CPU_SCORE_BLOCK;
-extern int count_occ( int *array, int size, int min, int max )
+/** Get average of values in array @arr of length @len excluding the
+ * best/worst value (like in speed cubing). */
+static double get_avg(double *arr, int len)
 {
-    int i, count;
-    count = 0;
-    for ( i = 0; i < size; i++ )
-        if ( array[i] >= min && array[i] <= max )
-            count++;
-    return count;
-}
-void tetris_make_stat()
-{
-    int i;
-    int game_count = 50;
-    double total = 0;
-    int total_lines = 0;
-    double scores[1024];
-    int lines[1024];
-    Bowl *bowl = 0;
-    SDL_Event event;
-    int leave = 0;
-    FILE *file = 0;
-    
-    printf( "*****\n" );
-    
-    bowl = bowl_create( 0, 0, -1, -1, -1, -1, blocks, qmark, "Demo", 0 );
-    
-    /* reset counters */
-    total = 0; total_lines = 0;
-    memset( lines, 0, sizeof( lines ) );
-    memset( scores, 0, sizeof( scores ) );
-        
-    printf( "Computing: %3i %3i %3i %3i %3i %3i\n", CPU_SCORE_HOLE, CPU_SCORE_ALT, CPU_SCORE_LINE, CPU_SCORE_STEEP, CPU_SCORE_ABYSS, CPU_SCORE_BLOCK );
-        
-    for ( i = 0; i < game_count; i++ ) {
-        if ( SDL_PollEvent( &event ) && event.type == SDL_KEYUP && event.key.keysym.sym == SDLK_ESCAPE )
-            leave = 1;
-        bowl_quick_game( bowl, 1 );
-        lines[i] = bowl->lines;
-        total_lines += lines[i];
-        scores[i] = bowl->score.value;
-        total += scores[i];
-        printf( "%3i: %5i: %14.0f\n", i, lines[i], scores[i] );
-        if ( leave ) return;
-    }
-    
-    if ( i != game_count ) game_count = i;
-    if ( game_count <= 0 ) return;
-    
-    /* write to file */
-    file = fopen( "stats", "a+" );
-    fprintf( file, "SETTING: %3i %3i %3i %3i %3i %3i\n", CPU_SCORE_HOLE, CPU_SCORE_ALT, CPU_SCORE_LINE, CPU_SCORE_STEEP, CPU_SCORE_ABYSS, CPU_SCORE_BLOCK );
-    for (i = 0; i < game_count;i++)
-        fprintf( file, "%3i: %5i: %14.0f\n", i, lines[i], scores[i] );
-    fprintf( file, "-----\n" );
-    fprintf( file, "  0-100: %4i  101-200: %4i  201-400: %4i\n", 
-             count_occ( lines, game_count, 0, 100 ), count_occ( lines, game_count, 101, 200 ), count_occ( lines, game_count, 201, 400 ) );
-    fprintf( file, "401-600: %4i  601-800: %4i  rest:    %4i\n", 
-             count_occ( lines, game_count, 401, 600 ), count_occ( lines, game_count, 601, 800 ), count_occ( lines, game_count, 801, 100000 ) );
-    fprintf( file, "-----\n" );
-    fprintf( file, "Avg.Lines: %i Avg.Score: %i\n", total_lines / game_count, (int)(total / game_count) );
-    fprintf( file, "\n" );
-    fclose( file );
-    
-    bowl_delete( bowl );
-}
-#ifdef _1
-void tetris_make_stat()
-{
-    int i;
-    int game_count = 10;
-    double total = 0;
-    int total_lines = 0;
-    int scores[1024];
-    int lines[1024];
-    Bowl *bowl = 0;
-    int ms = 4;
-    SDL_Event event;
-    int leave = 0;
-    char str[256];
-    
-    font->align = ALIGN_X_LEFT | ALIGN_Y_TOP;
-    tetris_recreate_bkgnd( 0 );
-    DEST( sdl.screen, 300, 0, 320, 480 ); fill_surf( 0x0 );
-    refresh_screen( 0, 0, 0, 0 );
-    if ( config.visualize )
-        shrapnells_init();
-    
-    for ( i = 0; i < game_count; i++ ) {
-        bowl = bowl_create( 0, 0, -1, -1, blocks, qmark, "Demo", 0 );
-        if ( config.visualize )
-            bowl->blind = 0;
-        else
-            bowl->blind = 1;
-        bowl->mute = 1;
-        while ( !bowl->game_over && !leave ) {
-            if ( SDL_PollEvent( &event ) && event.type == SDL_KEYUP && event.key.keysym.sym == SDLK_ESCAPE )
-                leave = 1;
-            bowl_update( bowl, ms, 0 );
-            if ( config.visualize )
-                refresh_rects();
-        }
-        lines[i] = bowl->lines;
-        scores[i] = counter_get_target( bowl->score );
-        sprintf( str, "%4i\n", lines[i] );
-        write_text( font, sdl.screen, 310 + ( i % 7 ) * 42, 10 + ( i / 7 ) * 16, str, OPAQUE );
-        refresh_screen( 0, 0, 0, 0 );
-        total += scores[i];
-        total_lines += bowl->lines;
-        bowl_delete( bowl );
-        printf( "%3i: %12i\n", i, scores[i] ); /* DEBUG */
-        if ( leave ) break;
-    }
-    
-    if ( config.visualize ) 
-        shrapnells_delete();
-    
-    if ( i != game_count )
-        game_count = i;
-    if ( game_count <= 0 ) return;
-    
-    sprintf( str, "Avg Score: %i Avg. Lines: %i\n", (int)(total / game_count), total_lines / game_count );
-    write_text( font, sdl.screen, 330, 428, str, OPAQUE );
-    
-    sprintf( str, "0-200: %i 201-400:: %i\n", count_occ( lines, game_count, 0, 200 ), count_occ( lines, game_count, 201, 400 ) );
-    write_text( font, sdl.screen, 330, 444, str, OPAQUE );
-    sprintf( str, "401-600: %i >601: %i\n", count_occ( lines, game_count, 401, 600 ), count_occ( lines, game_count, 601, 10000 ) );
-    write_text( font, sdl.screen, 330, 460, str, OPAQUE );
+	double min = arr[0], max = arr[0];
+	double sum = 0, count = 0;
 
-    refresh_screen( 0, 0, 0, 0 );
-        
-    wait_for_click();
+	/* get best and worst value */
+	for (int i = 0; i < len; i++) {
+		if (arr[i] < min)
+			min = arr[i];
+		if (arr[i] > max)
+			max = arr[i];
+	}
+
+	/* average values excluding best/worst */
+	for (int i = 0; i < len; i++) {
+		/* only exclude worst game for now, since we have a line cap
+		 * so many games have the same largest line count */
+		if (arr[i] > min/* && arr[i] < max*/) {
+			sum += arr[i];
+			count++;
+		}
+	}
+
+	if (count == 0)
+		return 0;
+	else
+		return sum / count;
+}
+
+/** Get average of all games that exceed line limit @llimit. */
+static double get_max_avg(double *lines, double *scores, int len, int llimit) {
+	double sum = 0, count = 0;
+
+	/* average values excluding best/worst */
+	for (int i = 0; i < len; i++) {
+		if (lines[i] >= llimit) {
+			sum += scores[i];
+			count++;
+		}
+	}
+
+	if (count == 0)
+		return 0;
+	else
+		return sum / count;
+}
+
+/** Run a single game. */
+double tetris_test_cpu_single(Bowl *bowl, CPU_ScoreSet *bscores, int verbose)
+{
+	int llimit = 2500;
+	int numgames = 5;
+	int dotval = numgames/20; /* show dot every this number of games */
+	double scores[numgames];
+	double lines[numgames];
+
+	/* clear counters */
+	memset(lines, 0, sizeof( lines ));
+	memset(scores, 0, sizeof( scores ));
+
+	printf( "Evaluating: l=%d h=%d s=%d a=%d b=%d c=%d\n",
+			bscores->lines, bscores->holes, bscores->slope,
+			bscores->abyss, bscores->block, bscores->clear);
+	if (!verbose)
+		printf("  ");
+
+	for (int i = 0; i < numgames; i++) {
+		/* check if quit requested */
+		SDL_PumpEvents();
+		unsigned char *keys = SDL_GetKeyState(NULL);
+		if (keys[SDLK_ESCAPE])
+			return 0;
+
+		/* run silent game and store results */
+		bowl_quick_game(bowl, bscores, llimit);
+		lines[i] = bowl->lines;
+		scores[i] = bowl->score.value;
+		if (verbose) {
+			printf( "  %3d: lines=%.0f, score=%.0f clears=[%d,%d,%d,%d]\n",
+						i, lines[i], scores[i],
+						bowl->stats.cleared[0],
+						bowl->stats.cleared[1],
+						bowl->stats.cleared[2],
+						bowl->stats.cleared[3]);
+		}
+		else if ((i % dotval) == 0) {
+			printf(".");
+			fflush(stdout);
+		}
+	}
+	printf("  -> avg: lines=%.0f, score=%.0f\n",
+		get_avg(lines,numgames), get_avg(scores, numgames));
+	printf("  avg score of maxed out games: %.0f\n",
+			get_max_avg(lines, scores, numgames, llimit));
+
+	return get_avg(scores, numgames);
+}
+
+/** Test CPU algorithm by running a number of games and printing some stats.
+ * type=1 - single run
+ * type=2 - test range of values */
+void tetris_test_cpu_algorithm(int type)
+{
+	Bowl *bowl = 0;
+	CPU_ScoreSet bscores;
+	CPU_ScoreSet bestset; /* result with highest game score */
+	double score, maxscore = 0;
+
+	printf( "**********\n" );
+	printf( "modern=%d, style=%d\n", config.modern, config.cpu_style);
+
+	bowl = bowl_create( 0, 0, -1, -1, -1, -1, blocks, qmark, "Demo", 0 );
+	if (config.modern) {
+		bowl->hold_active = 1;
+		bowl->hold_id = -1;
+	}
+
+	if (config.cpu_style == CS_AGGRESSIVE) {
+		/* slightly more aggressive play
+		 * results 1000 games, 5000 lines, modern=0:
+		 * 4250 avg lines, 53,5m avg score [66,7m max game avg score] */
+		bscores.lines = 15;
+		bscores.holes = -28;
+		bscores.slope = -2;
+		bscores.abyss = -7;
+		bscores.block = -5;
+		bscores.clear = 16;
+	} else {
+		/* normal settings, results 1000 games, 5000 lines, modern=0:
+		 * 4350 avg lines, 54,9m avg score [62,7m max game avg score] */
+		bscores.lines = 13;
+		bscores.holes = -28;
+		bscores.slope = -2;
+		bscores.abyss = -7;
+		bscores.block = -4;
+		bscores.clear = 16;
+	}
+
+	if (type == 1)
+		tetris_test_cpu_single(bowl, &bscores, 1);
+	else {
+		/* test variations of base scores */
+		//for (int i = 14; i <= 16; i++)
+		//for (int j = -29; j <= -27; j++)
+		//for (int k = 13; k <= 17; k++)
+		//for (int l = 12; l <= 13; l++)
+		for (int m = 12; m <= 28; m+=4) {
+			//bscores.lines = l;
+			//bscores.holes = l;
+			//bscores.slope = k;
+			//bscores.abyss = l;
+			//bscores.block = m;
+			bscores.clear = m;
+
+			score = tetris_test_cpu_single(bowl, &bscores, 0);
+
+			if (score > maxscore) {
+				bestset = bscores;
+				maxscore = score;
+			}
+		}
+
+		printf( "Best result: score=%0.f for l=%d h=%d s=%d a=%d b=%d, c=%d\n",
+				maxscore, bestset.lines, bestset.holes,
+				bestset.slope, bestset.abyss, bestset.block,
+				bscores.clear);
+	}
+
+	bowl_delete(bowl);
 }
-#endif
